from __future__ import annotations

import json
from enum import Enum
from typing import Any, Union

from ..core.data_model import DataModel


def _str_value(value: Any) -> Any:
    if value is None or isinstance(value, (str, int, float, bool, dict, list)):
        return json.dumps(value)
    return str(value)


class Parameter(DataModel):
    """Parameter.

    Attributes:
        name: Parameter name.
    """

    name: str

    def __str__(self) -> str:
        return f"@{self.name}"


class Ref(DataModel):
    """Parameter.

    Attributes:
        path: Ref path.
    """

    path: str

    def __str__(self) -> str:
        return f"{{{{{self.path}}}}}"

    @staticmethod
    def parse(str: str) -> Ref:
        if str.startswith("{{") and str.endswith("}}"):
            return Ref(path=str[2:-2])
        raise ValueError("Ref format error")


class FunctionNamespace:
    """Function namespace.

    Attributes:
        BUILTIN: Builtin namespace.
    """

    BUILTIN = "builtin"


class Function(DataModel):
    """Function.

    Attributes:
        name: Function name.
        args: Function args.
        named_args: Function named args.
        namespace: Function namespace.
    """

    namespace: str = FunctionNamespace.BUILTIN
    name: str
    args: list = []
    named_args: dict = dict()

    def __str__(self) -> str:
        if self.namespace == FunctionNamespace.BUILTIN:
            str = f"{self.name}"
        else:
            str = f"{self.namespace}.{self.name}"
        str = str + self._str_args()
        return str

    def _str_args(self):
        if len(self.args) > 0:
            return f"({', '.join(_str_value(a) for a in self.args)})"
        elif len(self.named_args) > 0:
            str_args = []
            for k, v in self.named_args.items():
                str_args.append(f"{k}={_str_value(v)}")
            return f"({', '.join(a for a in str_args)})"
        else:
            return "()"


class Undefined(DataModel):
    """Undefined."""

    pass


class Field(DataModel):
    """Field.

    Attributes:
        path: Field path.
    """

    path: str

    def __str__(self) -> str:
        return self.path


class Collection(DataModel):
    """Collection.

    Attributes:
        name: Collection name.
    """

    name: str

    def __str__(self) -> str:
        return self.name


class Comparison(DataModel):
    """Comparison expression.

    Attributes:
        lexpr: Left expression.
        op: Comparison op.
        rexpr: Right expression.
    """

    lexpr: Expression
    op: ComparisonOp
    rexpr: Expression

    def __str__(self) -> str:
        str = f"{_str_value(self.lexpr)} {self.op.value} "
        if self.op == ComparisonOp.BETWEEN and isinstance(self.rexpr, list):
            str = f"""{str}{_str_value(
                self.rexpr[0])} AND {_str_value(self.rexpr[1])}"""
        elif (
            self.op == ComparisonOp.IN or self.op == ComparisonOp.NIN
        ) and isinstance(self.rexpr, list):
            str = f"""{str}({', '.join(_str_value(i) for i in self.rexpr)})"""
        else:
            str = str + _str_value(self.rexpr)
        return str

    @staticmethod
    def reverse_op(op: ComparisonOp) -> ComparisonOp:
        if op == ComparisonOp.GT:
            return ComparisonOp.LT
        if op == ComparisonOp.GTE:
            return ComparisonOp.LTE
        if op == ComparisonOp.LT:
            return ComparisonOp.GT
        if op == ComparisonOp.LTE:
            return ComparisonOp.GTE
        return op


class ComparisonOp(str, Enum):
    """Comparison op.

    Attributes:
        LT: Less than.
        LTE: Less than equals.
        GT: Greater than.
        GTE: Greater than equals.
        EQ: Equals.
        NEQ: Not equals.
        IN: In.
        NIN: Not in.
        BETWEEN: Between.
        LIKE: Like.
    """

    LT = "<"
    LTE = "<="
    GT = ">"
    GTE = ">="
    EQ = "="
    NEQ = "!="
    IN = "in"
    NIN = "not in"
    BETWEEN = "between"
    LIKE = "like"


class And(DataModel):
    """And expression.

    Attributes:
        lexpr: Left expression.
        rexpr: Right expression.
    """

    lexpr: Expression
    rexpr: Expression

    def __str__(self) -> str:
        return f"({_str_value(self.lexpr)} AND {_str_value(self.rexpr)})"


class Or(DataModel):
    """Or expression.

    Attributes:
        lexpr: Left expression.
        rexpr: Right expression.
    """

    lexpr: Expression
    rexpr: Expression

    def __str__(self) -> str:
        return f"({_str_value(self.lexpr)} OR {_str_value(self.rexpr)})"


class Not(DataModel):
    """Not expression.

    Attributes:
        expr: Expression.
    """

    expr: Expression

    def __str__(self) -> str:
        return f"NOT {_str_value(self.expr)}"


class Update(DataModel):
    """Update.

    Attributes:
        operations: Update operations.
    """

    operations: list[UpdateOperation] = []

    def put(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.PUT,
                args=[value],
            )
        )
        return self

    def insert(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.INSERT,
                args=[value],
            )
        )
        return self

    def delete(self, field: str) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.DELETE,
            )
        )
        return self

    def increment(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.INCREMENT,
                args=[value],
            )
        )
        return self

    def move(self, field: str, dest: str) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.MOVE,
                args=[dest],
            )
        )
        return self

    def array_union(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.ARRAY_UNION,
                args=[value],
            )
        )
        return self

    def array_remove(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.ARRAY_REMOVE,
                args=[value],
            )
        )
        return self

    def append(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.APPEND,
                args=[value],
            )
        )
        return self

    def prepend(self, field: str, value: Expression) -> Update:
        self.operations.append(
            UpdateOperation(
                field=field,
                op=UpdateOp.PREPEND,
                args=[value],
            )
        )
        return self

    def __str__(self) -> str:
        return f"{', '.join(f'{str(o)}' for o in self.operations)}"


class UpdateOperation(DataModel):
    """Update operation.

    Attributes:
        field: Update field.
        op: Update op.
        args: Update args.
    """

    field: str
    op: UpdateOp
    args: list = []

    def __str__(self) -> str:
        str_args = None
        if self.args is not None:
            str_args = f"{', '.join(_str_value(a) for a in self.args)}"
        if self.args is not None:
            return f"{self.field}={self.op.value}({str_args})"
        else:
            return f"{self.field}={self.op.value}()"


class UpdateOp(str, Enum):
    """Update op.

    Attributes:
        PUT:
            Added if it doesn't exist.
            Replace if it exists.
            Array element replaced at index.
        INSERT:
            Added if it doesn't exist.
            Replace if it exists.
            Array element added at index.
            Use index - to insert at the end.
        DELETE:
            Removes field.
            Array element removed at index.
        INCREMENT:
            Increment or decrement number.
            Creates field if it doesn't exist.
        MOVE:
            Move value from source to destination field.
        ARRAY_UNION:
            Union values with array without duplicates.
        ARRAY_REMOVE:
            Remove values from array
        APPEND:
            Append existing string value.
        PREPEND:
            Prepend existing string value.

    """

    PUT = "put"
    INSERT = "insert"
    DELETE = "delete"
    INCREMENT = "increment"
    MOVE = "move"
    ARRAY_UNION = "array_union"
    ARRAY_REMOVE = "array_remove"
    APPEND = "append"
    PREPEND = "prepend"


class Select(DataModel):
    """Select.

    Attributes:
        terms: Select terms.
    """

    terms: list[SelectTerm] = []

    def add_field(self, field: str, alias: str | None = None):
        self.terms.append(SelectTerm(field=field, alias=alias))

    def __str__(self) -> str:
        if len(self.terms) == 0:
            return "*"
        return f"{', '.join(str(t) for t in self.terms)}"


class SelectTerm(DataModel):
    """Select term.

    Attributes:
        field: Select field.
        alias: Field alias.
    """

    field: str
    alias: str | None

    def __str__(self) -> str:
        str = self.field
        if self.alias is not None:
            str = f"{self.field} AS {self.alias}"
        return str


class OrderBy(DataModel):
    """Order by.

    Attributes:
        terms: Order by terms.
    """

    terms: list[OrderByTerm] = []

    def add_field(
        self,
        field: str,
        direction: OrderByDirection | None = None,
    ) -> OrderBy:
        self.terms.append(OrderByTerm(field=field, direction=direction))
        return self

    def __str__(self) -> str:
        return ", ".join([str(t) for t in self.terms])


class OrderByTerm(DataModel):
    """Order by term.

    Attributes:
        field: Order by field.
        direction: Order by direction.
    """

    field: str
    direction: OrderByDirection | None = None

    def __str__(self) -> str:
        str = self.field
        if self.direction:
            str = f"{str} {self.direction.value}"
        return str


class OrderByDirection(str, Enum):
    """Order by direction.

    Attributes:
        ASC: Ascending.
        DESC: Descending.
    """

    ASC = "asc"
    DESC = "desc"


class GeoPoint(DataModel):
    """Geo point.

    Attributes:
        lat: Latitude.
        lon: Longitude.
    """

    lat: float
    lon: float

    def __str__(self) -> str:
        return f"POINT({self.lon} {self.lat})"


Value = Union[str, int, float, bool, dict, list, bytes, DataModel, None]
Expression = Union[
    Comparison, And, Or, Not, Function, Parameter, Ref, Field, Value
]



import json
from typing import Any

from x0.core import Provider


class GoogleProvider(Provider):
    service_account_info: str | None
    service_account_file: str | None
    access_token: str | None

    _credentials: Any

    def __init__(
        self,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        **kwargs,
    ):
        self.service_account_info = service_account_info
        self.service_account_file = service_account_file
        self.access_token = access_token
        self._credentials = None
        super().__init__(**kwargs)

    def _get_credentials(self):
        if self._credentials:
            return self._credentials
        from google.oauth2 import service_account

        if self.service_account_info is not None:
            service_account_info = (
                json.loads(self.service_account_info)
                if isinstance(self.service_account_info, str)
                else self.service_account_info
            )
            self._credentials = (
                service_account.Credentials.from_service_account_info(
                    service_account_info
                )
            )
        elif self.service_account_file is not None:
            self._credentials = (
                service_account.Credentials.from_service_account_file(
                    self.service_account_file
                )
            )
        elif self.access_token is not None:
            from google.auth.credentials import Credentials

            self._credentials = Credentials(self.access_token)
        return self._credentials

    def _get_project_or_default(self, project: str | None) -> str:
        if project:
            return project
        return self._get_default_project()

    def _get_default_project(self) -> str:
        import google.auth

        _, project = google.auth.default()
        return project



from typing import Any

from ..ql._models import (
    And,
    Comparison,
    Function,
    Not,
    Or,
    Parameter,
    Update,
    UpdateOperation,
)
from ._operation import Operation
from .exceptions import BadRequestError


class ArgParser:
    parsable_args: dict = {
        "statement": "statement",
        "select": "select",
        "collection": "collection",
        "set": "update",
        "search": "search",
        "where": "where",
        "order_by": "order_by",
    }

    @staticmethod
    def convert_execute_operation(
        statement: str, params: dict[str, Any] | None
    ) -> Operation | None:
        operation = ArgParser.get_parsed_arg("statement", statement)
        if operation is None:
            return operation
        operation = ArgParser.replace_operation_params(operation, params)
        if params is not None:
            operation.args["params"] = params
        return operation

    @staticmethod
    def get_parsed_arg(
        arg_name: str,
        arg_value: Any,
        params: dict[str, Any] | None = None,
    ) -> Any:
        if arg_name in ArgParser.parsable_args:
            type = ArgParser.parsable_args[arg_name]
            if not isinstance(arg_value, str):
                return arg_value
            else:
                from verse.ql import QLParser

                parsed_arg = QLParser.parse(arg_value, type)
                if arg_name == "set":
                    parsed_arg = ArgParser.replace_set_params(
                        parsed_arg,
                        params,
                    )
                elif arg_name in ["search", "where"]:
                    parsed_arg = ArgParser.replace_expr_params(
                        parsed_arg, params
                    )
                return parsed_arg
        else:
            return arg_value

    @staticmethod
    def replace_operation_params(
        operation: Operation, params: dict[str, Any] | None
    ) -> Operation:
        if params is None or len(params) == 0:
            return operation
        if operation.args is not None:
            args: dict = {}
            for key, value in operation.args.items():
                args[key] = ArgParser.replace_expr_params(value, params)
            return Operation(name=operation.name, args=args)
        return Operation(name=operation.name)

    @staticmethod
    def replace_set_params(
        set: Update, params: dict[str, Any] | None
    ) -> Update:
        if params is None or len(params) == 0:
            return set
        copy = Update()
        for operation in set.operations:
            copy.operations.append(
                UpdateOperation(
                    field=operation.field,
                    op=operation.op,
                    args=[
                        ArgParser.replace_expr_params(arg, params)
                        for arg in operation.args
                    ],
                )
            )
        return copy

    @staticmethod
    def replace_expr_params(expr, params: dict[str, Any] | None) -> Any:
        if params is None or len(params) == 0:
            return expr
        if isinstance(expr, list):
            return [ArgParser.replace_expr_params(i, params) for i in expr]
        if isinstance(expr, Parameter):
            if expr.name in params:
                return params[expr.name]
            else:
                raise BadRequestError(f"Parameter {expr.name} not found")
        if isinstance(expr, Comparison):
            return Comparison(
                lexpr=ArgParser.replace_expr_params(expr.lexpr, params),
                op=expr.op,
                rexpr=ArgParser.replace_expr_params(expr.rexpr, params),
            )

        if isinstance(expr, Function):
            args = None
            named_args = None
            if expr.args is not None:
                args = [
                    ArgParser.replace_expr_params(arg, params)
                    for arg in expr.args
                ]
            if expr.named_args is not None:
                named_args = dict()
                for k, v in expr.named_args.items():
                    named_args[k] = ArgParser.replace_expr_params(v, params)

            return Function(
                namespace=expr.namespace,
                name=expr.name,
                args=args,
                named_args=named_args,
            )
        if isinstance(expr, And):
            return And(
                lexpr=ArgParser.replace_expr_params(expr.lexpr, params),
                rexpr=ArgParser.replace_expr_params(expr.rexpr, params),
            )
        if isinstance(expr, Or):
            return Or(
                lexpr=ArgParser.replace_expr_params(expr.lexpr, params),
                rexpr=ArgParser.replace_expr_params(expr.rexpr, params),
            )
        if isinstance(expr, Not):
            return Not(expr=ArgParser.replace_expr_params(expr.expr, params))
        return expr



from typing import Any

from uxp.core import Provider
from uxp.core.exceptions import BadRequestError


class AzureProvider(Provider):
    credential_type: str | None
    tenant_id: str | None
    client_id: str | None
    client_secret: str | None
    certificate_path: str | None

    _credential: Any
    _acredential: Any

    def __init__(
        self,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        **kwargs,
    ):
        self.credential_type = credential_type
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.certificate_path = certificate_path
        self._credential = None
        self._acredential = None
        super().__init__(**kwargs)

    def _get_credential(self):
        if self._credential:
            return self._credential
        credential_type = self.credential_type
        credential = None
        if credential_type == "default":
            from azure.identity import DefaultAzureCredential

            credential = DefaultAzureCredential()

        elif credential_type == "client_secret":
            from azure.identity import ClientSecretCredential

            credential = ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )

        elif credential_type == "certificate":
            from azure.identity import CertificateCredential

            credential = CertificateCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                certificate_path=self.certificate_path,
            )

        elif credential_type == "azure_cli":
            from azure.identity import AzureCliCredential

            credential = AzureCliCredential()

        elif credential_type == "shared_token_cache":
            from azure.identity import SharedTokenCacheCredential

            credential = SharedTokenCacheCredential()

        elif credential_type == "managed_identity":
            from azure.identity import ManagedIdentityCredential

            credential = ManagedIdentityCredential()

        self._credential = credential
        return credential

    def _aget_credential(self):
        if self._acredential:
            return self._acredential
        credential_type = self.credential_type
        credential = None
        if credential_type == "default":
            from azure.identity.aio import DefaultAzureCredential

            credential = DefaultAzureCredential()

        if credential_type == "client_secret":
            from azure.identity.aio import ClientSecretCredential

            credential = ClientSecretCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
            )

        if credential_type == "certificate":
            from azure.identity.aio import CertificateCredential

            credential = CertificateCredential(
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                certificate_path=self.certificate_path,
            )

        if credential_type == "azure_cli":
            from azure.identity.aio import AzureCliCredential

            credential = AzureCliCredential()

        if credential_type == "shared_token_cache":
            from azure.identity.aio import SharedTokenCacheCredential

            credential = SharedTokenCacheCredential()

        if credential_type == "managed_identity":
            from azure.identity.aio import ManagedIdentityCredential

            credential = ManagedIdentityCredential()

        self._acredential = credential
        return credential

    def _get_default_subscription_id(self) -> str:
        from azure.mgmt.resource import SubscriptionClient

        sub_client = SubscriptionClient(credential=self._get_credential())
        subs = sub_client.subscriptions.list()
        if not subs:
            raise BadRequestError(
                "No subscriptions found for the given credentials."
            )
        sub = next(sub_client.subscriptions.list())
        return sub.subscription_id

    def _ensure_resource_group(
        self,
        resource_group: str,
        location: str = "westus2",
        subscription_id: str | None = None,
    ) -> None:
        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.resource import ResourceManagementClient

        rg_client = ResourceManagementClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        try:
            rg_client.resource_groups.get(resource_group)
        except ResourceNotFoundError:
            rg_client.resource_groups.create_or_update(
                resource_group, {"location": location}
            )
            print(f"Created resource group: {resource_group}")

    def _delete_resource_group_if_empty(
        self,
        resource_group: str,
        subscription_id: str | None = None,
    ) -> None:
        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.resource import ResourceManagementClient

        rg_client = ResourceManagementClient(
            credential=self._credential,
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )

        # Check if the resource group is empty before deleting
        resources = list(
            rg_client.resources.list_by_resource_group(resource_group)
        )
        for resource in resources:
            print(f"Found resource in group: {resource.name}")
        if resources:
            print(
                f"Resource group {resource_group} is not empty; "
                "skipping deletion."
            )
            return
        try:
            operation = rg_client.resource_groups.begin_delete(resource_group)
            operation.result()
            print(f"Deleted resource group: {resource_group}")
        except ResourceNotFoundError:
            print(f"Resource group {resource_group} not found to delete.")



from typing import Any

from ._arg_parser import ArgParser
from ._async_helper import run_async, run_sync
from ._context import Context
from ._operation import Operation
from ._type_converter import TypeConverter
from .exceptions import NotSupportedError


class Provider:
    __component__: Any
    __handle__: str | None
    __type__: str

    def __init__(self, **kwargs):
        self.__handle__ = kwargs.pop("__handle__", None)
        self.__type__ = kwargs.pop("__type__", self.__class__.__module__)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __setup__(self, context: Context | None = None) -> None:
        pass

    async def __asetup__(self, context: Context | None = None) -> None:
        await run_async(func=self.__setup__, context=context)

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if operation and operation.name:
            func_name = operation.name
            func = getattr(self, func_name, None)
            if func and callable(func):
                self.__setup__(context=context)
                args = TypeConverter.convert_args(func, operation.args or {})
                response = func(**args)
                return response

            afunc_name = f"a{operation.name}"
            afunc = getattr(self, afunc_name, None)
            if callable(afunc):
                run_sync(self.__asetup__, context=context)
                args = TypeConverter.convert_args(afunc, operation.args or {})
                response = run_sync(afunc, **args)
                return response
        raise NotSupportedError(
            operation.to_json() if operation is not None else None
        )

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        if operation and operation.name:
            afunc_name = f"a{operation.name}"
            afunc = getattr(self, afunc_name, None)
            if afunc and callable(afunc):
                await self.__asetup__(context=context)
                args = TypeConverter.convert_args(afunc, operation.args or {})
                response = await afunc(**args)
                return response

        return await run_async(
            func=self.__run__,
            operation=operation,
            context=context,
            **kwargs,
        )

    def __supports__(self, feature: str) -> bool:
        return True

    def __execute__(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        context: Context | None = None,
        **kwargs: Any,
    ) -> Any:
        operation = ArgParser.convert_execute_operation(statement, params)
        return self.__run__(operation, context)

    async def __aexecute__(
        self,
        statement: str,
        params: dict[str, Any] | None = None,
        context: Context | None = None,
        **kwargs: Any,
    ) -> Any:
        operation = ArgParser.convert_execute_operation(statement, params)
        return await self.__arun__(operation, context)



# flake8: noqa
# type: ignore
# Generated from VerseQLParser.g4 by ANTLR 4.13.2
from antlr4 import *

if "." in __name__:
    from .VerseQLParser import VerseQLParser
else:
    from VerseQLParser import VerseQLParser


# This class defines a complete listener for a parse tree produced by VerseQLParser.
class VerseQLParserListener(ParseTreeListener):

    # Enter a parse tree produced by VerseQLParser#identifier.
    def enterIdentifier(self, ctx: VerseQLParser.IdentifierContext):
        pass

    # Exit a parse tree produced by VerseQLParser#identifier.
    def exitIdentifier(self, ctx: VerseQLParser.IdentifierContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_statement.
    def enterParse_statement(self, ctx: VerseQLParser.Parse_statementContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_statement.
    def exitParse_statement(self, ctx: VerseQLParser.Parse_statementContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_search.
    def enterParse_search(self, ctx: VerseQLParser.Parse_searchContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_search.
    def exitParse_search(self, ctx: VerseQLParser.Parse_searchContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_where.
    def enterParse_where(self, ctx: VerseQLParser.Parse_whereContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_where.
    def exitParse_where(self, ctx: VerseQLParser.Parse_whereContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_select.
    def enterParse_select(self, ctx: VerseQLParser.Parse_selectContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_select.
    def exitParse_select(self, ctx: VerseQLParser.Parse_selectContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_collection.
    def enterParse_collection(
        self, ctx: VerseQLParser.Parse_collectionContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_collection.
    def exitParse_collection(self, ctx: VerseQLParser.Parse_collectionContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_order_by.
    def enterParse_order_by(self, ctx: VerseQLParser.Parse_order_byContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_order_by.
    def exitParse_order_by(self, ctx: VerseQLParser.Parse_order_byContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_rank_by.
    def enterParse_rank_by(self, ctx: VerseQLParser.Parse_rank_byContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_rank_by.
    def exitParse_rank_by(self, ctx: VerseQLParser.Parse_rank_byContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parse_update.
    def enterParse_update(self, ctx: VerseQLParser.Parse_updateContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parse_update.
    def exitParse_update(self, ctx: VerseQLParser.Parse_updateContext):
        pass

    # Enter a parse tree produced by VerseQLParser#statement_single.
    def enterStatement_single(
        self, ctx: VerseQLParser.Statement_singleContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#statement_single.
    def exitStatement_single(self, ctx: VerseQLParser.Statement_singleContext):
        pass

    # Enter a parse tree produced by VerseQLParser#statement_multi.
    def enterStatement_multi(self, ctx: VerseQLParser.Statement_multiContext):
        pass

    # Exit a parse tree produced by VerseQLParser#statement_multi.
    def exitStatement_multi(self, ctx: VerseQLParser.Statement_multiContext):
        pass

    # Enter a parse tree produced by VerseQLParser#clause.
    def enterClause(self, ctx: VerseQLParser.ClauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#clause.
    def exitClause(self, ctx: VerseQLParser.ClauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#generic_clause.
    def enterGeneric_clause(self, ctx: VerseQLParser.Generic_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#generic_clause.
    def exitGeneric_clause(self, ctx: VerseQLParser.Generic_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#select_clause.
    def enterSelect_clause(self, ctx: VerseQLParser.Select_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_clause.
    def exitSelect_clause(self, ctx: VerseQLParser.Select_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#select_all.
    def enterSelect_all(self, ctx: VerseQLParser.Select_allContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_all.
    def exitSelect_all(self, ctx: VerseQLParser.Select_allContext):
        pass

    # Enter a parse tree produced by VerseQLParser#select_terms.
    def enterSelect_terms(self, ctx: VerseQLParser.Select_termsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_terms.
    def exitSelect_terms(self, ctx: VerseQLParser.Select_termsContext):
        pass

    # Enter a parse tree produced by VerseQLParser#select_parameter.
    def enterSelect_parameter(
        self, ctx: VerseQLParser.Select_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#select_parameter.
    def exitSelect_parameter(self, ctx: VerseQLParser.Select_parameterContext):
        pass

    # Enter a parse tree produced by VerseQLParser#select_term.
    def enterSelect_term(self, ctx: VerseQLParser.Select_termContext):
        pass

    # Exit a parse tree produced by VerseQLParser#select_term.
    def exitSelect_term(self, ctx: VerseQLParser.Select_termContext):
        pass

    # Enter a parse tree produced by VerseQLParser#collection_clause.
    def enterCollection_clause(
        self, ctx: VerseQLParser.Collection_clauseContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_clause.
    def exitCollection_clause(
        self, ctx: VerseQLParser.Collection_clauseContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#collection_identifier.
    def enterCollection_identifier(
        self, ctx: VerseQLParser.Collection_identifierContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_identifier.
    def exitCollection_identifier(
        self, ctx: VerseQLParser.Collection_identifierContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#collection_parameter.
    def enterCollection_parameter(
        self, ctx: VerseQLParser.Collection_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#collection_parameter.
    def exitCollection_parameter(
        self, ctx: VerseQLParser.Collection_parameterContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#search_clause.
    def enterSearch_clause(self, ctx: VerseQLParser.Search_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#search_clause.
    def exitSearch_clause(self, ctx: VerseQLParser.Search_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#where_clause.
    def enterWhere_clause(self, ctx: VerseQLParser.Where_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#where_clause.
    def exitWhere_clause(self, ctx: VerseQLParser.Where_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_operand.
    def enterExpression_operand(
        self, ctx: VerseQLParser.Expression_operandContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_operand.
    def exitExpression_operand(
        self, ctx: VerseQLParser.Expression_operandContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_not.
    def enterExpression_not(self, ctx: VerseQLParser.Expression_notContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_not.
    def exitExpression_not(self, ctx: VerseQLParser.Expression_notContext):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_or.
    def enterExpression_or(self, ctx: VerseQLParser.Expression_orContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_or.
    def exitExpression_or(self, ctx: VerseQLParser.Expression_orContext):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_comparison.
    def enterExpression_comparison(
        self, ctx: VerseQLParser.Expression_comparisonContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison.
    def exitExpression_comparison(
        self, ctx: VerseQLParser.Expression_comparisonContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_comparison_in.
    def enterExpression_comparison_in(
        self, ctx: VerseQLParser.Expression_comparison_inContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison_in.
    def exitExpression_comparison_in(
        self, ctx: VerseQLParser.Expression_comparison_inContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_paranthesis.
    def enterExpression_paranthesis(
        self, ctx: VerseQLParser.Expression_paranthesisContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_paranthesis.
    def exitExpression_paranthesis(
        self, ctx: VerseQLParser.Expression_paranthesisContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_comparison_between.
    def enterExpression_comparison_between(
        self, ctx: VerseQLParser.Expression_comparison_betweenContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_comparison_between.
    def exitExpression_comparison_between(
        self, ctx: VerseQLParser.Expression_comparison_betweenContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#expression_and.
    def enterExpression_and(self, ctx: VerseQLParser.Expression_andContext):
        pass

    # Exit a parse tree produced by VerseQLParser#expression_and.
    def exitExpression_and(self, ctx: VerseQLParser.Expression_andContext):
        pass

    # Enter a parse tree produced by VerseQLParser#operand_value.
    def enterOperand_value(self, ctx: VerseQLParser.Operand_valueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_value.
    def exitOperand_value(self, ctx: VerseQLParser.Operand_valueContext):
        pass

    # Enter a parse tree produced by VerseQLParser#operand_field.
    def enterOperand_field(self, ctx: VerseQLParser.Operand_fieldContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_field.
    def exitOperand_field(self, ctx: VerseQLParser.Operand_fieldContext):
        pass

    # Enter a parse tree produced by VerseQLParser#operand_parameter.
    def enterOperand_parameter(
        self, ctx: VerseQLParser.Operand_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_parameter.
    def exitOperand_parameter(
        self, ctx: VerseQLParser.Operand_parameterContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#operand_ref.
    def enterOperand_ref(self, ctx: VerseQLParser.Operand_refContext):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_ref.
    def exitOperand_ref(self, ctx: VerseQLParser.Operand_refContext):
        pass

    # Enter a parse tree produced by VerseQLParser#operand_function.
    def enterOperand_function(
        self, ctx: VerseQLParser.Operand_functionContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#operand_function.
    def exitOperand_function(self, ctx: VerseQLParser.Operand_functionContext):
        pass

    # Enter a parse tree produced by VerseQLParser#order_by_clause.
    def enterOrder_by_clause(self, ctx: VerseQLParser.Order_by_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_clause.
    def exitOrder_by_clause(self, ctx: VerseQLParser.Order_by_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#order_by_terms.
    def enterOrder_by_terms(self, ctx: VerseQLParser.Order_by_termsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_terms.
    def exitOrder_by_terms(self, ctx: VerseQLParser.Order_by_termsContext):
        pass

    # Enter a parse tree produced by VerseQLParser#order_by_parameter.
    def enterOrder_by_parameter(
        self, ctx: VerseQLParser.Order_by_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_parameter.
    def exitOrder_by_parameter(
        self, ctx: VerseQLParser.Order_by_parameterContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#order_by_term.
    def enterOrder_by_term(self, ctx: VerseQLParser.Order_by_termContext):
        pass

    # Exit a parse tree produced by VerseQLParser#order_by_term.
    def exitOrder_by_term(self, ctx: VerseQLParser.Order_by_termContext):
        pass

    # Enter a parse tree produced by VerseQLParser#rank_by_clause.
    def enterRank_by_clause(self, ctx: VerseQLParser.Rank_by_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#rank_by_clause.
    def exitRank_by_clause(self, ctx: VerseQLParser.Rank_by_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#set_clause.
    def enterSet_clause(self, ctx: VerseQLParser.Set_clauseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#set_clause.
    def exitSet_clause(self, ctx: VerseQLParser.Set_clauseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#update_operations.
    def enterUpdate_operations(
        self, ctx: VerseQLParser.Update_operationsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_operations.
    def exitUpdate_operations(
        self, ctx: VerseQLParser.Update_operationsContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#update_parameter.
    def enterUpdate_parameter(
        self, ctx: VerseQLParser.Update_parameterContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_parameter.
    def exitUpdate_parameter(self, ctx: VerseQLParser.Update_parameterContext):
        pass

    # Enter a parse tree produced by VerseQLParser#update_operation.
    def enterUpdate_operation(
        self, ctx: VerseQLParser.Update_operationContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#update_operation.
    def exitUpdate_operation(self, ctx: VerseQLParser.Update_operationContext):
        pass

    # Enter a parse tree produced by VerseQLParser#function.
    def enterFunction(self, ctx: VerseQLParser.FunctionContext):
        pass

    # Exit a parse tree produced by VerseQLParser#function.
    def exitFunction(self, ctx: VerseQLParser.FunctionContext):
        pass

    # Enter a parse tree produced by VerseQLParser#function_no_args.
    def enterFunction_no_args(
        self, ctx: VerseQLParser.Function_no_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_no_args.
    def exitFunction_no_args(self, ctx: VerseQLParser.Function_no_argsContext):
        pass

    # Enter a parse tree produced by VerseQLParser#function_with_args.
    def enterFunction_with_args(
        self, ctx: VerseQLParser.Function_with_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_with_args.
    def exitFunction_with_args(
        self, ctx: VerseQLParser.Function_with_argsContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#function_with_named_args.
    def enterFunction_with_named_args(
        self, ctx: VerseQLParser.Function_with_named_argsContext
    ):
        pass

    # Exit a parse tree produced by VerseQLParser#function_with_named_args.
    def exitFunction_with_named_args(
        self, ctx: VerseQLParser.Function_with_named_argsContext
    ):
        pass

    # Enter a parse tree produced by VerseQLParser#named_arg.
    def enterNamed_arg(self, ctx: VerseQLParser.Named_argContext):
        pass

    # Exit a parse tree produced by VerseQLParser#named_arg.
    def exitNamed_arg(self, ctx: VerseQLParser.Named_argContext):
        pass

    # Enter a parse tree produced by VerseQLParser#ref.
    def enterRef(self, ctx: VerseQLParser.RefContext):
        pass

    # Exit a parse tree produced by VerseQLParser#ref.
    def exitRef(self, ctx: VerseQLParser.RefContext):
        pass

    # Enter a parse tree produced by VerseQLParser#ref_path.
    def enterRef_path(self, ctx: VerseQLParser.Ref_pathContext):
        pass

    # Exit a parse tree produced by VerseQLParser#ref_path.
    def exitRef_path(self, ctx: VerseQLParser.Ref_pathContext):
        pass

    # Enter a parse tree produced by VerseQLParser#parameter.
    def enterParameter(self, ctx: VerseQLParser.ParameterContext):
        pass

    # Exit a parse tree produced by VerseQLParser#parameter.
    def exitParameter(self, ctx: VerseQLParser.ParameterContext):
        pass

    # Enter a parse tree produced by VerseQLParser#field.
    def enterField(self, ctx: VerseQLParser.FieldContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field.
    def exitField(self, ctx: VerseQLParser.FieldContext):
        pass

    # Enter a parse tree produced by VerseQLParser#field_path.
    def enterField_path(self, ctx: VerseQLParser.Field_pathContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field_path.
    def exitField_path(self, ctx: VerseQLParser.Field_pathContext):
        pass

    # Enter a parse tree produced by VerseQLParser#field_primitive.
    def enterField_primitive(self, ctx: VerseQLParser.Field_primitiveContext):
        pass

    # Exit a parse tree produced by VerseQLParser#field_primitive.
    def exitField_primitive(self, ctx: VerseQLParser.Field_primitiveContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_null.
    def enterValue_null(self, ctx: VerseQLParser.Value_nullContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_null.
    def exitValue_null(self, ctx: VerseQLParser.Value_nullContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_true.
    def enterValue_true(self, ctx: VerseQLParser.Value_trueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_true.
    def exitValue_true(self, ctx: VerseQLParser.Value_trueContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_false.
    def enterValue_false(self, ctx: VerseQLParser.Value_falseContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_false.
    def exitValue_false(self, ctx: VerseQLParser.Value_falseContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_string.
    def enterValue_string(self, ctx: VerseQLParser.Value_stringContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_string.
    def exitValue_string(self, ctx: VerseQLParser.Value_stringContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_integer.
    def enterValue_integer(self, ctx: VerseQLParser.Value_integerContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_integer.
    def exitValue_integer(self, ctx: VerseQLParser.Value_integerContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_decimal.
    def enterValue_decimal(self, ctx: VerseQLParser.Value_decimalContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_decimal.
    def exitValue_decimal(self, ctx: VerseQLParser.Value_decimalContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_json.
    def enterValue_json(self, ctx: VerseQLParser.Value_jsonContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_json.
    def exitValue_json(self, ctx: VerseQLParser.Value_jsonContext):
        pass

    # Enter a parse tree produced by VerseQLParser#value_array.
    def enterValue_array(self, ctx: VerseQLParser.Value_arrayContext):
        pass

    # Exit a parse tree produced by VerseQLParser#value_array.
    def exitValue_array(self, ctx: VerseQLParser.Value_arrayContext):
        pass

    # Enter a parse tree produced by VerseQLParser#literal_string.
    def enterLiteral_string(self, ctx: VerseQLParser.Literal_stringContext):
        pass

    # Exit a parse tree produced by VerseQLParser#literal_string.
    def exitLiteral_string(self, ctx: VerseQLParser.Literal_stringContext):
        pass

    # Enter a parse tree produced by VerseQLParser#array_empty.
    def enterArray_empty(self, ctx: VerseQLParser.Array_emptyContext):
        pass

    # Exit a parse tree produced by VerseQLParser#array_empty.
    def exitArray_empty(self, ctx: VerseQLParser.Array_emptyContext):
        pass

    # Enter a parse tree produced by VerseQLParser#array_items.
    def enterArray_items(self, ctx: VerseQLParser.Array_itemsContext):
        pass

    # Exit a parse tree produced by VerseQLParser#array_items.
    def exitArray_items(self, ctx: VerseQLParser.Array_itemsContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json.
    def enterJson(self, ctx: VerseQLParser.JsonContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json.
    def exitJson(self, ctx: VerseQLParser.JsonContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_obj.
    def enterJson_obj(self, ctx: VerseQLParser.Json_objContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_obj.
    def exitJson_obj(self, ctx: VerseQLParser.Json_objContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_pair.
    def enterJson_pair(self, ctx: VerseQLParser.Json_pairContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_pair.
    def exitJson_pair(self, ctx: VerseQLParser.Json_pairContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_arr.
    def enterJson_arr(self, ctx: VerseQLParser.Json_arrContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_arr.
    def exitJson_arr(self, ctx: VerseQLParser.Json_arrContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_value.
    def enterJson_value(self, ctx: VerseQLParser.Json_valueContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_value.
    def exitJson_value(self, ctx: VerseQLParser.Json_valueContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_string.
    def enterJson_string(self, ctx: VerseQLParser.Json_stringContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_string.
    def exitJson_string(self, ctx: VerseQLParser.Json_stringContext):
        pass

    # Enter a parse tree produced by VerseQLParser#json_number.
    def enterJson_number(self, ctx: VerseQLParser.Json_numberContext):
        pass

    # Exit a parse tree produced by VerseQLParser#json_number.
    def exitJson_number(self, ctx: VerseQLParser.Json_numberContext):
        pass


del VerseQLParser



import inspect
from functools import wraps
from typing import Any, Callable, TypeVar, cast

from ._operation import Operation
from .exceptions import NotSupportedError

T = TypeVar("T", bound=Callable[..., Any])


def component(**config: Any) -> Callable[[T], T]:
    def decorator(cls: T) -> T:
        setattr(cls, "__config__", config)
        return cls

    return decorator


def operation(**config: Any) -> Callable[[T], T]:
    def decorator(func: T) -> T:
        setattr(func, "__operation__", True)
        setattr(func, "__config__", config)
        if not inspect.iscoroutinefunction(func):

            @wraps(func)
            def wrapper(*args, **kwargs) -> Any:
                self = args[0]
                context = kwargs.pop("__context__", None)
                if hasattr(self, "__provider__"):
                    sig = inspect.signature(func)
                    bound_args = sig.bind(*args, **kwargs)
                    bound_args.apply_defaults()
                    locals = dict(bound_args.arguments)
                    locals.pop("self", None)
                    operation = Operation.normalize(
                        name=func.__name__,
                        args=locals,
                    )
                    try:
                        response = self.__run__(operation, context, **kwargs)
                        return response
                    except NotSupportedError:
                        return func(*args, **kwargs)
                else:
                    return func(*args, **kwargs)

            return cast(T, wrapper)
        else:

            @wraps(func)
            async def wrapper(*args, **kwargs) -> Any:
                self = args[0]
                context = kwargs.pop("__context__", None)
                if hasattr(self, "__provider__"):
                    sig = inspect.signature(func)
                    bound_args = sig.bind(*args, **kwargs)
                    bound_args.apply_defaults()
                    locals = dict(bound_args.arguments)
                    locals.pop("self", None)
                    operation = Operation.normalize(
                        name=func.__name__[1:],
                        args=locals,
                    )
                    try:
                        response = await self.__arun__(
                            operation, context, **kwargs
                        )
                        return response
                    except NotSupportedError:
                        return await func(*args, **kwargs)
                else:
                    return await func(*args, **kwargs)

            return cast(T, wrapper)

    return decorator



from typing import Any, Generic, TypeVar

from ._context import Context
from .data_model import DataModel

T = TypeVar("T")


class Response(DataModel, Generic[T]):
    result: T
    """Result of the primary provider."""

    context: Context | None = None
    """Operation context."""

    native: dict[str, Any] | None = None
    """Logs from native client."""



from __future__ import annotations

import copy
import importlib
import importlib.util
import inspect
import os
import re
import sys
from enum import Enum
from typing import Any

from ._component import Component, Provider
from ._log_helper import warn
from ._operation import Operation
from ._type_converter import TypeConverter
from .constants import ROOT_PACKAGE_NAME
from .dependency import (
    ComponentEdge,
    ComponentNode,
    ProviderEdge,
    ProviderNode,
)
from .exceptions import LoadError
from .manifest import MANIFEST_FILE, Manifest


class Loader:
    path: str
    manifest_path: str

    manifest: Manifest

    _init = False
    _component_dependencies: dict[str, ComponentDep] = dict()
    _provider_dependencies: dict[str, ProviderDep] = dict()

    def __init__(
        self,
        path: str = ".",
        manifest: str = MANIFEST_FILE,
    ):
        self.path = path
        self.manifest_path = os.path.join(
            path,
            manifest,
        )
        self.manifest = Manifest.parse(path=self.manifest_path)
        abs_project_folder = os.path.abspath(path)
        if abs_project_folder not in sys.path:
            sys.path.append(abs_project_folder)

    def get_component_type(self, handle: str) -> str:
        if handle in self.manifest.components:
            return self.manifest.components[handle].type
        raise LoadError("Unresolved handle")

    def load_component(
        self,
        handle: str,
        tag: str | None = None,
    ) -> Component:
        component = self._resolve_component(
            handle=handle,
            tag=tag,
        ).component
        return component

    def _resolve_component(
        self,
        handle: str,
        tag: str | None,
    ) -> ComponentDep:
        if handle not in self.manifest.components:
            raise LoadError(f"Component {handle} not found in manifest")
        cconfig = self.manifest.components.get(handle)
        if handle in self._component_dependencies:
            if self._component_dependencies[handle].resolved:
                return self._component_dependencies[handle]
            cdep = self._component_dependencies[handle]
        else:
            cdep = self._prepare_component_dependency(
                handle,
                cconfig.type,
                cconfig.parameters,
                cconfig.requirements,
            )
            self._component_dependencies[handle] = cdep
        cdep.parameters = self._resolve_param(
            cdep.parameters,
            tag,
        )
        if cconfig.providers:
            phandle = self._resolve_provider_handle(
                chandle=handle,
                tag=tag,
            )
            pdep = self._resolve_provider(
                chandle=handle,
                phandle=phandle,
                tag=tag,
            )
            cdep.provider = pdep
            provider = pdep.provider
        else:
            provider = None
        cdep.component = self._init_component(cdep, provider)
        cdep.component.__handle__ = handle
        cdep.component.__type__ = cdep.type
        cdep.resolved = True
        return cdep

    def _resolve_provider_handle(
        self,
        chandle: str,
        tag: str | None = None,
    ) -> str:
        if chandle not in self.manifest.components:
            raise LoadError(f"Component {chandle} not found in manifest")
        cconfig = self.manifest.components[chandle]
        if tag and tag in self.manifest.bindings:
            bindings = self.manifest.bindings[tag]
            if chandle in bindings:
                phandle = bindings[chandle]
                if phandle not in cconfig.providers:
                    raise LoadError(
                        (
                            f"Provider handle {phandle} not found in "
                            f"providers for {chandle}"
                        )
                    )
                return phandle
        return next(iter(cconfig.providers.keys()))

    def _resolve_provider(
        self,
        chandle: str,
        phandle: str,
        tag: str | None = None,
    ) -> ProviderDep:
        if chandle not in self.manifest.components:
            raise LoadError(f"Component {chandle} not found in manifest")
        if phandle not in self.manifest.components[chandle].providers:
            raise LoadError(
                f"Provider {phandle} not found in component " f"{chandle}"
            )
        handle = f"{chandle}:{phandle}"
        if handle in self._provider_dependencies:
            if self._provider_dependencies[handle].resolved:
                return self._provider_dependencies[handle]
            pdep = self._provider_dependencies[handle]
        else:
            cconfig = self.manifest.components[chandle]
            pconfig = cconfig.providers[phandle]
            pdep = self._prepare_provider_dependency(
                handle=phandle,
                type=pconfig.type,
                parameters=pconfig.parameters,
                requirements=pconfig.requirements,
                component_type=cconfig.type,
            )
            self._provider_dependencies[handle] = pdep
        pdep.parameters = self._resolve_param(
            pdep.parameters,
            tag,
        )
        pdep.provider = self._init_provider(pdep)
        pdep.provider.__handle__ = pdep.handle
        pdep.provider.__type__ = pdep.type
        pdep.resolved = True
        return pdep

    def _resolve_param(self, value: Any, tag: str | None) -> Any:
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = self._resolve_param(value=v, tag=tag)
            return value
        elif isinstance(value, list):
            for i, item in enumerate(value):
                value[i] = self._resolve_param(value=item, tag=tag)
            return value
        elif isinstance(value, str) and self._is_ref(value):
            ref_info = self._get_ref_info(value, tag)
            return self._resolve_ref(ref_info=ref_info, tag=tag)
        return value

    def _init_component(
        self,
        cdep: ComponentDep,
        provider: Provider | None,
    ) -> Component:
        path = cdep.py_path
        component = Loader.load_component_instance(
            path=path,
            parameters=cdep.parameters,
            provider=provider,
        )
        return component

    def _init_provider(self, pdep: ProviderDep) -> Provider:
        path = pdep.py_path
        provider = Loader.load_provider_instance(
            path=path,
            parameters=pdep.parameters,
        )
        return provider

    def _prepare_component_dependency(
        self,
        handle: str,
        type: str,
        parameters: dict[str, Any],
        requirements: list[str],
    ) -> ComponentDep:
        cdep = ComponentDep()
        cdep.handle = handle
        cdep.type = type
        cdep.parameters = copy.deepcopy(parameters)
        cdep.requirements = requirements
        cdep.py_path = Loader.get_component_path(component_type=type)
        cdep.kind = Loader.get_component_kind(component_type=type)
        cdep.resolved = False
        return cdep

    def _prepare_provider_dependency(
        self,
        handle: str,
        type: str,
        parameters: dict[str, Any],
        requirements: list[str],
        component_type: str,
    ) -> ProviderDep:
        pdep = ProviderDep()
        pdep.handle = handle
        pdep.component_type = type
        pdep.type = type
        pdep.parameters = copy.deepcopy(parameters)
        pdep.requirements = requirements
        pdep.py_path = Loader.get_provider_path(
            component_type=component_type,
            provider_type=type,
        )
        pdep.kind = Loader.get_provider_kind(
            component_type=component_type,
            provider_type=type,
        )
        pdep.resolved = False
        return pdep

    def _resolve_ref(self, ref_info: RefInfo, tag: str | None) -> Any:
        if ref_info.type == RefType.CONTEXT:
            return self._resolve_context(
                param=ref_info.param,
                tag=tag,
            )
        if ref_info.type == RefType.ENV:
            return self._resolve_param(
                value=self._resolve_env(ref_info.param),
                tag=tag,
            )
        if ref_info.type == RefType.VARIABLE:
            return self._resolve_param(
                value=self._resolve_variable(ref_info.param),
                tag=tag,
            )
        if ref_info.type == RefType.METADATA:
            return self._resolve_metadata(
                param=ref_info.param,
            )
        if ref_info.type == RefType.COMPONENT:
            return self._resolve_component(
                handle=ref_info.component_handle,
                tag=tag,
            ).component
        elif ref_info.type == RefType.COMPONENT_GET:
            return self._resolve_component_get(
                self._resolve_component(
                    handle=ref_info.component_handle,
                    tag=tag,
                ).component,
                ref_info.get_args,
            )
        elif ref_info.type == RefType.COMPONENT_PARAM:
            return self._resolve_component_param(
                self._resolve_component(
                    handle=ref_info.component_handle,
                    tag=tag,
                ).component,
                ref_info.param,
            )
        elif ref_info.type == RefType.PROVIDER:
            return self._resolve_provider(
                chandle=ref_info.component_handle,
                phandle=ref_info.provider_handle,
                tag=tag,
            ).provider
        elif ref_info.type == RefType.PROVIDER_GET:
            return self._resolve_provider_get(
                self._resolve_provider(
                    chandle=ref_info.component_handle,
                    phandle=ref_info.provider_handle,
                    tag=tag,
                ).provider,
                ref_info.get_args,
            )
        elif ref_info.type == RefType.PROVIDER_PARAM:
            return self._resolve_provider_param(
                self._resolve_provider(
                    chandle=ref_info.component_handle,
                    phandle=ref_info.provider_handle,
                    tag=tag,
                ).provider,
                ref_info.param,
            )

    def _resolve_context(self, param: str, tag: str | None) -> Any:
        context = {"tag": tag}
        if param in context:
            return context[param]
        return None

    def _resolve_env(self, param: str) -> Any:
        value = os.getenv(param)
        return value

    def _resolve_variable(self, param: str) -> Any:
        if param in self.manifest.variables:
            return self.manifest.variables[param]
        return None

    def _resolve_metadata(self, param: str) -> Any:
        if param in self.manifest.metadata:
            return self.manifest.metadata[param]
        return None

    def _resolve_component_param(
        self, component: Component, param: str
    ) -> Any:
        return getattr(component, param, None)

    def _resolve_provider_param(self, provider: Provider, param: str) -> Any:
        return getattr(provider, param, None)

    def _resolve_component_get(self, component: Component, get_args: dict):
        res = component.__run__(operation=Operation(name="get", args=get_args))
        return res.result.value

    def _resolve_provider_get(self, provider: Provider, get_args: dict):
        res = provider.__run__(operation=Operation(name="get", args=get_args))
        return res.result.value

    def _resolve_selector(
        self,
        component_handle: str,
        selector: str,
        tag: str | None,
    ) -> str:
        if component_handle not in self.manifest.components:
            raise LoadError(f"Component handle {component_handle} not found")
        binding_key = f"{component_handle}.{selector}"
        if self.manifest.bindings:
            if tag:
                if tag not in self.manifest.bindings:
                    raise LoadError(f"Tag {tag} not found in bindings")
                bindings = self.manifest.bindings.get(tag)
            else:
                for key, value in self.manifest.bindings.items():
                    bindings = value
                    warn(
                        "Warning: No tag provided. "
                        f"Selecting the first set of bindings {key}."
                    )
                    break
            if binding_key in bindings:
                return bindings[binding_key]
        selected_provider = next(
            iter(self.manifest.components[component_handle].providers.keys())
        )
        warn(
            f"Warning: Selector binding {binding_key} not found for "
            f"{component_handle}. Using provider {selected_provider}."
        )
        return selected_provider

    def _get_ref_info(self, string: str, tag: str | None) -> RefInfo:
        def parse_get_args(key: str) -> dict:
            parts = key.split(".")
            return (
                {"key": {"label": parts[0], "id": parts[1]}}
                if len(parts) == 2
                else {"key": {"id": key}}
            )

        def split_and_clean(value: str) -> tuple[str, str]:
            return value.split("[")[0], value.split("[")[1].split("]")[
                0
            ].replace("'", "")

        match = re.match(r"^\$\{(.+)\}$", string)
        if not match:
            raise LoadError(f"{string} not a ref")

        value = match.group(1)
        ref_info = RefInfo()
        ref_info.value = value
        prefix_map = {
            "env.": RefType.ENV,
            "context.": RefType.CONTEXT,
            "metadata.": RefType.METADATA,
            "variables.": RefType.VARIABLE,
        }

        for prefix, ref_type in prefix_map.items():
            if value.startswith(prefix):
                ref_info.type = ref_type
                start = len(prefix)
                ref_info.param = value[start:]
                return ref_info

        if "[" in value and "]" in value:
            base, key = split_and_clean(value)
            ref_info.get_args = parse_get_args(key)

            if "::" in base:
                ref_info.type = RefType.PROVIDER_GET
                comp, sel = base.split("::")
                ref_info.component_handle = comp
                ref_info.provider_handle = self._resolve_selector(
                    comp, sel, tag
                )
            elif ":" in base:
                ref_info.type = RefType.PROVIDER_GET
                ref_info.component_handle, ref_info.provider_handle = (
                    base.split(":")
                )
            else:
                ref_info.type = RefType.COMPONENT_GET
                ref_info.component_handle = base
            return ref_info

        if "::" in value:
            comp, rest = value.split("::")
            ref_info.component_handle = comp
            if "." in rest:
                ref_info.type = RefType.PROVIDER_PARAM
                sel, param = rest.split(".")
                ref_info.provider_handle = self._resolve_selector(
                    comp, sel, tag
                )
                ref_info.param = param
            else:
                ref_info.type = RefType.PROVIDER
                ref_info.provider_handle = self._resolve_selector(
                    comp, rest, tag
                )
            return ref_info

        if ":" in value:
            comp, rest = value.split(":")
            ref_info.component_handle = comp
            if "." in rest:
                ref_info.type = RefType.PROVIDER_PARAM
                ref_info.provider_handle, ref_info.param = rest.split(".")
            else:
                ref_info.type = RefType.PROVIDER
                ref_info.provider_handle = rest
            return ref_info

        if "." in value:
            ref_info.type = RefType.COMPONENT_PARAM
            ref_info.component_handle, ref_info.param = value.split(".")
            return ref_info

        ref_info.type = RefType.COMPONENT
        ref_info.component_handle = value
        return ref_info

    def _is_ref(self, str: str) -> bool:
        pattern = r"^\$\{(.+)\}$"
        return bool(re.match(pattern, str))

    def _get_composite_component_provider(
        self,
        type: str,
    ) -> tuple[str, str]:
        _, package = type.split(".", 1)
        if "." in package:
            component, provider = package.split(".", 1)
        else:
            component = package
            provider = "default"
        return component, provider

    def generate_requirements(
        self,
        handle: str,
        tag: str | None = None,
        out: str | None = None,
    ) -> dict[str, list[str]]:
        base_requirements: dict[str, list[str]] = {
            "verse-core": [],
            "verse-content": [],
        }
        custom_requirements: set[str] = set()

        def traverse_provider(
            cnode: ComponentNode, pnode: ProviderNode | None
        ):
            if pnode is None:
                return
            if pnode.requirements:
                custom_requirements.update(pnode.requirements)
            else:
                if pnode.kind == "composite":
                    package = "verse-composite"
                    component, provider = (
                        self._get_composite_component_provider(pnode.type)
                    )
                elif pnode.kind == "base":
                    type = cnode.type.replace(f"{ROOT_PACKAGE_NAME}.", "")
                    cnamespace, cpackage = type.split(".", 1)
                    package = f"verse-{cnamespace}"
                    component = cpackage
                    provider = pnode.type
                else:
                    package = None
                if package:
                    extra = (
                        component.replace("_", "-")
                        + "-"
                        + provider.replace("_", "-")
                    )
                    if package not in base_requirements:
                        base_requirements[package] = []
                    if extra not in base_requirements[package]:
                        base_requirements[package].append(extra)

            for child in pnode.depends:
                traverse_component(child.component)
            for child in pnode.composes:
                traverse_provider(cnode=cnode, pnode=child.provider)

        def traverse_component(cnode: ComponentNode):
            custom_requirements.update(cnode.requirements)
            traverse_provider(cnode=cnode, pnode=cnode.provider)
            for child in cnode.depends:
                traverse_component(child.component)

        custom_requirements.update(self.manifest.requirements)
        graph = self.get_dependency_graph(handle=handle, tag=tag)
        traverse_component(graph)

        reqs: list[str] = []
        for package in base_requirements:
            if len(base_requirements[package]) > 0:
                reqs.append(
                    package
                    + "["
                    + str.join(",", base_requirements[package])
                    + "]"
                )
            else:
                reqs.append(package)
        if len(custom_requirements) > 0:
            for req in custom_requirements:
                reqs.append(req)
        if out is not None:
            with open(out, "w") as file:
                file.write("\n".join(reqs))
        return reqs

    def get_dependency_graph(
        self,
        handle: str,
        tag: str | None = None,
    ) -> ComponentNode:
        component_node_cache: dict[str, ComponentNode] = {}
        provider_node_cache: dict[str, ProviderNode] = {}

        def _get_component_node(cdep: ComponentDep) -> ComponentNode:
            if cdep.handle in component_node_cache:
                return component_node_cache[cdep.handle]
            node = ComponentNode(
                handle=cdep.handle,
                type=cdep.type,
                kind=cdep.kind,
                requirements=cdep.requirements,
            )
            if cdep.provider:
                node.provider = _get_provider_node(
                    cdep=cdep, pdep=cdep.provider
                )
            component_dependencies = _get_component_dependencies_from_param(
                value=cdep.parameters, tag=tag
            )
            for cdep, ref_info in component_dependencies:
                node.depends.append(
                    ComponentEdge(
                        component=_get_component_node(cdep),
                        ref=ref_info.value,
                    )
                )
            component_node_cache[cdep.handle] = node
            return node

        def _get_provider_node(
            cdep: ComponentDep, pdep: ProviderDep
        ) -> ProviderNode:
            handle = f"{cdep.handle}:{pdep.handle}"
            if handle in provider_node_cache:
                return provider_node_cache[handle]
            node = ProviderNode(
                handle=pdep.handle,
                type=pdep.type,
                kind=pdep.kind,
                requirements=pdep.requirements,
            )
            component_dependencies = _get_component_dependencies_from_param(
                value=pdep.parameters, tag=tag
            )
            for component_dependency, ref_info in component_dependencies:
                node.depends.append(
                    ComponentEdge(
                        component=_get_component_node(component_dependency),
                        ref=ref_info.value,
                    )
                )
            provider_dependencies = _get_provider_dependencies_from_param(
                value=pdep.parameters, tag=tag
            )
            for pdep, ref_info in provider_dependencies:
                node.composes.append(
                    ProviderEdge(
                        provider=_get_provider_node(cdep=cdep, pdep=pdep),
                        ref=ref_info.value,
                    )
                )
            provider_node_cache[handle] = node
            return node

        def _get_provider_dependencies_from_param(
            value: Any,
            tag: str | None,
        ) -> list[tuple[ProviderDep, RefInfo]]:
            dependencies: list[ProviderDep, RefInfo] = []
            if isinstance(value, dict):
                for k, v in value.items():
                    dependencies.extend(
                        _get_provider_dependencies_from_param(value=v, tag=tag)
                    )
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    dependencies.extend(
                        _get_provider_dependencies_from_param(
                            value=item, tag=tag
                        )
                    )
            elif isinstance(value, str) and self._is_ref(value):
                ref_info = self._get_ref_info(value, tag)
                if ref_info.type in (
                    RefType.PROVIDER,
                    RefType.PROVIDER_PARAM,
                    RefType.PROVIDER_GET,
                ):
                    handle = (
                        f"{ref_info.component_handle}:"
                        f"{ref_info.provider_handle}"
                    )
                    if handle in self._provider_dependencies:
                        dependencies.append(
                            (self._provider_dependencies[handle], ref_info)
                        )
                    else:
                        raise LoadError(f"Unresolved provider {handle}")
            return dependencies

        def _get_component_dependencies_from_param(
            value: Any,
            tag: str | None,
        ) -> list[tuple[ComponentDep, RefInfo]]:
            dependencies: list = []
            if isinstance(value, dict):
                for k, v in value.items():
                    dependencies.extend(
                        _get_component_dependencies_from_param(
                            value=v, tag=tag
                        )
                    )
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    dependencies.extend(
                        _get_component_dependencies_from_param(
                            value=item, tag=tag
                        )
                    )
            elif isinstance(value, str) and self._is_ref(value):
                ref_info = self._get_ref_info(value, tag)
                if ref_info.type in (
                    RefType.COMPONENT,
                    RefType.COMPONENT_PARAM,
                    RefType.COMPONENT_GET,
                ):
                    handle = ref_info.component_handle
                    if handle in self._component_dependencies:
                        dependencies.append(
                            (self._component_dependencies[handle], ref_info)
                        )
                    else:
                        raise LoadError(f"Unresolved component {handle}")
            return dependencies

        def _prepare_unresolved(
            tag: str | None = None,
        ):
            for chandle in self.manifest.components:
                cconfig = self.manifest.components[chandle]
                cdep = self._prepare_component_dependency(
                    chandle,
                    cconfig.type,
                    cconfig.parameters,
                    cconfig.requirements,
                )
                self._component_dependencies[chandle] = cdep
                if cconfig.providers:
                    phandle = self._resolve_provider_handle(
                        chandle=chandle,
                        tag=tag,
                    )
                    pconfig = cconfig.providers[phandle]
                    pdep = self._prepare_provider_dependency(
                        handle=phandle,
                        type=pconfig.type,
                        parameters=pconfig.parameters,
                        requirements=pconfig.requirements,
                        component_type=cconfig.type,
                    )
                    cdep.provider = pdep
                    self._provider_dependencies[f"{chandle}:{phandle}"] = pdep
                else:
                    cdep.provider = None

        _prepare_unresolved(tag=tag)
        node = _get_component_node(self._component_dependencies[handle])
        return node

    @staticmethod
    def load_provider_instance(
        path: str | None = None,
        parameters: dict[str, Any] = dict(),
    ) -> Provider:
        if path is None:
            return Provider(**parameters)
        provider = Loader.load_class(path, Provider)
        converted_parameters = TypeConverter.convert_args(
            provider.__init__, parameters
        )
        return provider(**converted_parameters)

    @staticmethod
    def get_component_kind(
        component_type: str,
    ) -> str:
        if component_type.startswith(f"{ROOT_PACKAGE_NAME}."):
            return "base"
        return "custom"

    @staticmethod
    def get_provider_kind(
        component_type: str,
        provider_type: str,
    ) -> str:
        if provider_type.startswith(f"{ROOT_PACKAGE_NAME}.composite."):
            return "composite"
        elif component_type.startswith(f"{ROOT_PACKAGE_NAME}."):
            return "base"
        else:
            return "custom"

    @staticmethod
    def get_component_path(
        component_type: str,
    ) -> str:
        if ":" in component_type:
            return component_type
        else:
            return f"{component_type}.component"

    @staticmethod
    def get_provider_path(
        component_type: str,
        provider_type: str,
    ) -> str:
        if ":" in provider_type or ":" in component_type:
            return provider_type
        elif provider_type.startswith(f"{ROOT_PACKAGE_NAME}."):
            if ".providers." in provider_type:
                return provider_type
            else:
                return f"{provider_type}.providers.default"
        else:
            return f"{component_type}.providers.{provider_type}"

    @staticmethod
    def load_component_instance(
        path: str | None,
        parameters: dict,
        provider: Provider | None,
    ) -> Component:
        if path is None:
            return Component(__provider__=provider, **parameters)
        component = Loader.load_class(path, Component)
        converted_parameters = TypeConverter.convert_args(
            component.__init__, parameters
        )
        return component(__provider__=provider, **converted_parameters)

    @staticmethod
    def load_class(path: str, type: Any) -> Any:
        class_name = None
        if ":" in path:
            splits = path.split(":")
            module_name = splits[0]
            class_name = splits[-1]
        else:
            module_name = path
        if module_name.endswith(".py"):
            import sys

            normalized_module_name = (
                os.path.normpath(module_name)
                .replace("\\", "/")
                .split(".py")[0]
                .replace("/", ".")
                .lstrip(".")
            )
            spec = importlib.util.spec_from_file_location(
                normalized_module_name, os.path.abspath(path)
            )
            if spec is not None:
                module = importlib.util.module_from_spec(spec)
                sys.modules[normalized_module_name] = module
                if spec.loader is not None:
                    spec.loader.exec_module(module)
            module_name = normalized_module_name
        else:
            module = importlib.import_module(module_name)
        if class_name is not None:
            return getattr(module, class_name)
        else:
            for name, cls in inspect.getmembers(module, inspect.isclass):
                if issubclass(cls, type) and cls.__module__ == module_name:
                    return cls
        raise LoadError(f"{type.__name__} not found at {module}")


class RefInfo:
    type: RefType
    component_handle: str
    provider_handle: str
    param: str
    get_args: dict
    value: str


class RefType(str, Enum):
    ENV = "env"
    CONTEXT = "context"
    METADATA = "metadata"
    VARIABLE = "variable"
    COMPONENT = "component"
    COMPONENT_PARAM = "component_param"
    COMPONENT_GET = "component_get"
    PROVIDER = "provider"
    PROVIDER_PARAM = "provider_param"
    PROVIDER_GET = "provider_get"


class ComponentDep:
    handle: str
    type: str
    parameters: dict[str, Any]
    requirements: list[str]
    kind: str
    py_path: str
    provider: ProviderDep
    component: Component
    resolved: bool

    def __init__(self):
        self.parameters = dict()


class ProviderDep:
    handle: str
    component_type: str
    type: str
    parameters: dict[str, Any]
    requirements: list[str]
    kind: str
    py_path: str
    provider: Provider
    resolved: bool

    def __init__(self):
        self.parameters = dict()



from typing import Any

from ..ql._models import Field, Undefined, UpdateOp, Value
from .data_model import DataModel
from .exceptions import BadRequestError


class DataAccessor:
    @staticmethod
    def get_field(
        item: dict | DataModel | None, field: str
    ) -> Value | Undefined:
        if item is None:
            return Undefined()
        npath = (
            field.replace("[", "/")
            .replace("]", "")
            .replace(".", "/")
            .rstrip("/")
        )
        splits = npath.split("/")
        current_item = item
        for split in splits:
            if split.isnumeric() and isinstance(current_item, list):
                index = int(split)
                if index < len(current_item):
                    current_item = current_item[index]
                else:
                    return Undefined()
            elif split == "-":
                if isinstance(current_item, list):
                    if len(current_item) > 0:
                        current_item = current_item[-1]
                    else:
                        return Undefined()
                else:
                    return Undefined()
            else:
                if isinstance(current_item, DataModel):
                    if hasattr(current_item, split):
                        current_item = getattr(current_item, split)
                    else:
                        return Undefined()
                else:
                    if not isinstance(current_item, dict):
                        return Undefined()
                    if split in current_item:
                        current_item = current_item[split]
                    else:
                        return Undefined()
        return current_item

    @staticmethod
    def update_field(item: dict, field: str, op: UpdateOp, value: Any):
        npath = (
            field.replace("[", "/")
            .replace("]", "")
            .replace(".", "/")
            .rstrip("/")
        )
        splits = npath.split("/")
        current_item = item
        last_split = False
        len_splits = len(splits)
        for i in range(len_splits):
            split = splits[i]
            last_split = i == len_splits - 1
            if split.isnumeric() and isinstance(current_item, list):
                index = int(split)
                if not last_split:
                    if index < len(current_item):
                        current_item = current_item[index]
                    else:
                        raise BadRequestError(
                            f"Index {index} more than length of the array"
                        )
                else:
                    if op == UpdateOp.PUT:
                        if index >= 0 and index < len(current_item):
                            current_item[index] = value
                        else:
                            raise BadRequestError(
                                f"Index {index} is out of range"
                            )
                    elif op == UpdateOp.INSERT:
                        if index >= 0 and index <= len(current_item):
                            current_item.insert(index, value)
                        else:
                            raise BadRequestError(
                                f"Index {index} is out of range"
                            )
                    elif op == UpdateOp.DELETE:
                        if index >= 0 and index < len(current_item):
                            current_item.pop(index)
                        else:
                            raise BadRequestError(
                                f"Index {index} is out of range"
                            )
                    elif op == UpdateOp.INCREMENT:
                        if index >= 0 and index < len(current_item):
                            if isinstance(current_item[index], (int, float)):
                                current_item[index] += value
                            else:
                                raise BadRequestError(
                                    "Increment field should be a number"
                                )
                        else:
                            raise BadRequestError(
                                f"Index {index} is out of range"
                            )
                    else:
                        raise BadRequestError(
                            "Operation not supported on arrays"
                        )
            elif split == "-":
                if isinstance(current_item, list):
                    if not last_split:
                        raise BadRequestError(
                            "- can be used only as a suffix in field path"
                        )
                    else:
                        if op == UpdateOp.PUT:
                            if len(current_item) > 0:
                                current_item[-1] = value
                            else:
                                raise BadRequestError(
                                    "Array is empty for SET operation"
                                )
                        elif op == UpdateOp.INSERT:
                            current_item.append(value)
                        elif op == UpdateOp.DELETE:
                            if len(current_item) > 0:
                                current_item.pop(-1)
                            else:
                                raise BadRequestError(
                                    "Array is empty for DELETE operation"
                                )
                        else:
                            raise BadRequestError(
                                "Operation not supported on arrays"
                            )
            else:
                if not last_split:
                    if split not in current_item:
                        if isinstance(current_item, DataModel):
                            setattr(current_item, split, dict())
                        else:
                            current_item[split] = dict()
                    if isinstance(current_item, DataModel):
                        current_item = getattr(current_item, split)
                    else:
                        current_item = current_item[split]
                else:
                    if op == UpdateOp.PUT:
                        if isinstance(current_item, DataModel):
                            setattr(current_item, split, value)
                        else:
                            current_item[split] = value
                    elif op == UpdateOp.INSERT:
                        if isinstance(current_item, DataModel):
                            setattr(current_item, split, value)
                        else:
                            current_item[split] = value
                    elif op == UpdateOp.DELETE:
                        if isinstance(current_item, DataModel):
                            delattr(current_item, split)
                        else:
                            if split in current_item:
                                current_item.pop(split)
                    elif op == UpdateOp.INCREMENT:
                        if isinstance(current_item, DataModel):
                            setattr(
                                current_item,
                                split,
                                getattr(current_item, split) + value,
                            )
                        else:
                            if isinstance(current_item[split], (int, float)):
                                current_item[split] += value
                            else:
                                raise BadRequestError(
                                    "Increment field should be a number"
                                )
                    elif op == UpdateOp.ARRAY_UNION:
                        if isinstance(current_item, DataModel):
                            lst = getattr(current_item, split)
                            if isinstance(lst, list):
                                for arg in value:
                                    if arg not in lst:
                                        lst.append(arg)
                            else:
                                raise BadRequestError(
                                    f"Field {split} must be an array"
                                )
                        else:
                            if isinstance(current_item[split], list):
                                for arg in value:
                                    if arg not in current_item[split]:
                                        current_item[split].append(arg)
                            else:
                                raise BadRequestError(
                                    f"Field {split} must be an array"
                                )
                    elif op == UpdateOp.ARRAY_REMOVE:
                        if isinstance(current_item, DataModel):
                            lst = getattr(current_item, split)
                            if isinstance(lst, list):
                                for arg in value:
                                    if arg in lst:
                                        lst.remove(arg)
                            else:
                                raise BadRequestError(
                                    f"Field {split} must be an array"
                                )
                        else:
                            if isinstance(current_item[split], list):
                                for arg in value:
                                    if arg in current_item[split]:
                                        current_item[split].remove(arg)
                            else:
                                raise BadRequestError(
                                    f"Field {split} must be an array"
                                )
                    elif op == UpdateOp.MOVE:
                        present = False
                        if isinstance(current_item, DataModel):
                            present = hasattr(current_item, split)
                        else:
                            present = split in current_item
                        if present:
                            DataAccessor.update_field(
                                item,
                                (
                                    value.path
                                    if isinstance(value, Field)
                                    else value
                                ),
                                UpdateOp.PUT,
                                current_item[split],
                            )
                            if isinstance(current_item, DataModel):
                                delattr(current_item, split)
                            else:
                                current_item.pop(split)
                        else:
                            BadRequestError(f"Field {split} not found")
